#include "queues.h"
#include <asm/errno.h>
#include <time.h>
#include <stdio.h>


void init_queue(queue_s *q){
    q->mtx = (pthread_mutex_t *) malloc(sizeof(pthread_mutex_t));	
    q->cond = (pthread_cond_t *) malloc(sizeof(pthread_cond_t));	
    pthread_mutex_init(q->mtx,NULL);
    pthread_cond_init(q->cond,NULL);
    q->buflen = 1<<7;
    q->buf =(void **) malloc( q->buflen * sizeof(void *) );
    q->p = q->k = 0;
}

unsigned int queue_len(queue_s *q){
    return (q->p>q->k)?q->buflen-(q->p-q->k):q->k-q->p;
}

void queue_grow(queue_s *q){
    q->buf = (void **) realloc (q->buf,sizeof(void *) * (q->buflen<<1));
    if (q->p > q->k) {
	memmove( q->buf + q->p + q->buflen, q->buf + q->p, (q->buflen - q->p)*sizeof(void *));
	q->p += q->buflen;
    }
    q->buflen <<= 1;
}

void _queue_put(queue_s *q, void *p){
    if (((q->k+1)&(q->buflen-1)) == q->p) {
	queue_grow(q);
    }
    q->buf[q->k++] = p;
    q->k &= q->buflen-1;
}

void *_queue_get(queue_s *q){
    void *ret;
    
    if (q->k == q->p) return NULL;
    ret = q->buf[q->p++];
    q->p &= q->buflen-1;
    return ret;
}


void queue_put(queue_s *q, void *p){
    pthread_mutex_lock(q->mtx);
    _queue_put(q,p);
    pthread_cond_signal(q->cond);
    pthread_mutex_unlock(q->mtx);
}

void *queue_get(queue_s *q){
    void *ret;
    pthread_mutex_lock(q->mtx);
    ret = _queue_get(q);
    pthread_mutex_unlock(q->mtx);
    return ret;
}

void *queue_get_wait(queue_s *q){
    void *ret;
    pthread_mutex_lock(q->mtx);
    do {
	ret = _queue_get(q);
	if (ret==NULL) pthread_cond_wait(q->cond,q->mtx);
    } while (ret==NULL);
    pthread_mutex_unlock(q->mtx);
    return ret;
}

void *queue_get_timedwait(queue_s *q, int msecs){
    void *ret;
    struct timespec time = {0,msecs*1000000};
    int waitret;
    
    pthread_mutex_lock(q->mtx);
    do {
	ret = _queue_get(q);
	if (ret==NULL) waitret = pthread_cond_timedwait(q->cond,q->mtx,&time);
    } while (ret==NULL && waitret!=ETIMEDOUT);
    pthread_mutex_unlock(q->mtx);
    return ret;
}

void *queue_index(queue_s *q,unsigned int i){
    if (i>=queue_len(q)) return NULL;
    else return q->buf[(i+q->p)&(q->buflen-1)];
}

void queue_remove_index(queue_s *q,unsigned int i){
    int pos=(i+q->p)&(q->buflen-1);
    
    if (i>=queue_len(q)) return ;
    if (q->p > q->k) {
	if (pos >= q->p) {
	    memmove( q->buf + q->p + 1 , q->buf + q->p , (pos - q->p)*sizeof(void *));
	    q->p++;
	    if (q->p==q->buflen) q->p=0;
	} else {
	    memmove( q->buf + pos , q->buf + pos + 1 , (q->k - pos)*sizeof(void *));
	    q->k--;
	    if (q->k<0) q->k = q->buflen-1;
	}
    } else {
	memmove( q->buf + pos , q->buf + pos + 1 ,(q->k - pos -1)*sizeof(void *));	
	q->k--;
    }
}
